@@ -9,6 +9,7 @@ import polars as pl
 from kedro.config import OmegaConfigLoader
 from kedro.framework.project import settings
 from sqlalchemy import create_engine, text, tuple_
+from sqlalchemy import text as sa_text
 from sqlalchemy.dialects.postgresql import insert
 from sqlalchemy.engine import Engine
 from sqlalchemy.exc import SQLAlchemyError
@@ -343,7 +344,7 @@ class SQLAlchemyDatabase:
         data: pl.DataFrame,
         mode: t.Literal["RAISE", "IGNORE", "UPDATE"],
         *,
-        chunk_size: int = 100,
+        chunk_size: int = 10000,
         returning_id: bool = False,
     ) -> pl.DataFrame | None:
         """
@@ -356,70 +357,204 @@ class SQLAlchemyDatabase:
         """
         # Determine unique-key columns
         p_unique_key_columns = self.get_primary_unique_constraint(model_class)
-        total_rows = data.height
-        total_chunks = (total_rows + chunk_size - 1) // chunk_size
-
-        logger.info("Starting bulk insert of %d rows in %d chunks of size %d", total_rows, total_chunks, chunk_size)
-        # Phase 1: Write all rows in chunks
+        logger.info(
+            "Starting bulk insert of %d rows in %d chunks of size %d",
+            data.height,
+            (data.height + chunk_size - 1) // chunk_size,
+            chunk_size,
+        )
         table_name = model_class.__table__.name
-        with self.get_session() as session:
-            for offset in tqdm(
-                range(0, total_rows, chunk_size),
-                total=total_chunks,
-                desc=f"Writing to {table_name} with chunk size {chunk_size}",
-                unit="chunk",
-            ):
-                chunk_df = data.slice(offset, chunk_size)
-                records = chunk_df.to_dicts()
+        self._write_chunks(
+            model_class,
+            data,
+            mode,
+            chunk_size,
+            p_unique_key_columns,
+            table_name,
+        )
 
-                # Build INSERT statement for only the columns present in the DataFrame
-                stmt = insert(model_class.__table__).values({col: bindparam(col) for col in chunk_df.columns})
-                if mode == "IGNORE":
-                    stmt = stmt.on_conflict_do_nothing(index_elements=p_unique_key_columns)
-                elif mode == "UPDATE":
-                    update_cols = [c for c in chunk_df.columns if c not in p_unique_key_columns]
-                    if update_cols:
-                        stmt = stmt.on_conflict_do_update(
-                            index_elements=p_unique_key_columns, set_={col: stmt.excluded[col] for col in update_cols}
-                        )
-                    else:
-                        stmt = stmt.on_conflict_do_nothing(index_elements=p_unique_key_columns)
-                elif mode == "RAISE":
-                    pass  # No ON CONFLICT clause needed
-
-                # Execute insertion
-                session.execute(stmt, records)
-
-        # Phase 2: Retrieve IDs if needed
         if returning_id:
             logger.info("Fetching IDs for unique keys: %s", p_unique_key_columns)
-            # Collect unique-key tuples
-            key_df = data.select(*p_unique_key_columns)
-
-            # Query all matching IDs in one go
-            with self.get_session() as session:
-                # Map database column names to SQLAlchemy attributes
-                unique_key_attrs = [self.get_column_attr_by_db_name(model_class, col) for col in p_unique_key_columns]
-
-                query = session.query(
-                    *unique_key_attrs,
-                    model_class.id,
-                )
-                # filter by tuple IN
-                key_tuples = [tuple(row) if len(p_unique_key_columns) > 1 else row[0] for row in key_df.rows()]
-                if len(p_unique_key_columns) == 1:
-                    col_attr = unique_key_attrs[0]
-                    query = query.filter(col_attr.in_(key_tuples))
-                else:
-                    query = query.filter(tuple_(*unique_key_attrs).in_(key_tuples))
-                results = query.all()
-
-            # Build a Polars DataFrame of keys + id
-            ids_df = pl.DataFrame([dict(zip([*p_unique_key_columns, "id"], row, strict=True)) for row in results])
-            return ids_df
+            return self._fetch_ids_chunks(
+                model_class,
+                data,
+                chunk_size,
+                p_unique_key_columns,
+                table_name,
+            )
 
         return None
 
+    def _write_chunks(
+        self,
+        model_class: DeclarativeMeta,
+        data: pl.DataFrame,
+        mode: t.Literal["RAISE", "IGNORE", "UPDATE"],
+        chunk_size: int,
+        p_unique_key_columns: list[str],
+        table_name: str,
+    ) -> None:
+        """Helper to write data into the database in chunks."""
+        total_rows = data.height
+        total_chunks = (total_rows + chunk_size - 1) // chunk_size
+        row_iter = range(0, total_rows, chunk_size)
+        show_progress = total_chunks > 1
+        progress_bar = (
+            tqdm(
+                row_iter,
+                total=total_rows,
+                desc=f"Writing to table: {table_name}",
+                unit="rows",
+                dynamic_ncols=True,
+                mininterval=0.1,
+            )
+            if show_progress
+            else row_iter
+        )
+        columns = data.columns
+        stmt = insert(model_class.__table__).values({col: bindparam(col) for col in columns})
+        if mode == "IGNORE":
+            stmt = stmt.on_conflict_do_nothing(index_elements=p_unique_key_columns)
+        elif mode == "UPDATE":
+            update_cols = [c for c in columns if c not in p_unique_key_columns]
+            if update_cols:
+                stmt = stmt.on_conflict_do_update(
+                    index_elements=p_unique_key_columns, set_={col: stmt.excluded[col] for col in update_cols}
+                )
+            else:
+                stmt = stmt.on_conflict_do_nothing(index_elements=p_unique_key_columns)
+        elif mode == "RAISE":
+            pass  # No ON CONFLICT clause needed
+
+        with self.get_session() as session:
+            for i, offset in enumerate(row_iter):
+                chunk_df = data.slice(offset, chunk_size)
+                records = chunk_df.to_dicts()
+                session.execute(stmt, records)
+                # Manually update progress bar to show correct number of rows processed
+                if show_progress:
+                    # Calculate how many rows have been processed so far
+                    processed = min((i + 1) * chunk_size, total_rows)
+                    progress_bar.n = processed
+                    progress_bar.last_print_n = processed
+                    progress_bar.update(0)
+
+    def _fetch_ids_chunks(
+        self,
+        model_class: DeclarativeMeta,
+        data: pl.DataFrame,
+        chunk_size: int,
+        p_unique_key_columns: list[str],
+        table_name: str,
+        use_temp_table: int = 1000,
+        profile: bool = False,
+    ) -> pl.DataFrame:
+        """Helper to fetch IDs in chunks after writing data. Uses temp table for large chunks."""
+
+        def func(model_class, data, chunk_size, p_unique_key_columns, table_name, use_temp_table):
+            key_df = data.select(*p_unique_key_columns)
+            total_keys = key_df.height
+            total_chunks = (total_keys + chunk_size - 1) // chunk_size
+            row_iter = range(0, total_keys, chunk_size)
+            show_progress = total_chunks > 1
+            progress_bar = (
+                tqdm(
+                    row_iter,
+                    total=total_keys,
+                    desc=f"Fetching IDs from table: {table_name}",
+                    unit="rows",
+                    dynamic_ncols=True,
+                    mininterval=0.01,
+                )
+                if show_progress
+                else row_iter
+            )
+            unique_key_attrs = [self.get_column_attr_by_db_name(model_class, col) for col in p_unique_key_columns]
+            ids = []
+            use_temp = chunk_size > use_temp_table
+            with self.get_session() as session:
+                # Precompute invariants for temp table approach
+                if use_temp:
+                    # Get column types from model (once)
+                    col_defs = []
+                    for col in p_unique_key_columns:
+                        col_obj = model_class.__table__.columns[col]
+                        col_type = col_obj.type.compile(dialect=session.bind.dialect)
+                        col_defs.append(f'"{col}" {col_type}')
+                    col_defs_sql = ", ".join(col_defs)
+                    col_names_sql = ", ".join([f'"{col}"' for col in p_unique_key_columns])
+                    param_names = [f":{col}" for col in p_unique_key_columns]
+                # Get the actual primary key column name from the model (once)
+                pk_col = next(iter(model_class.__table__.primary_key)).name
+                # Precompute join condition and select columns for temp table join
+                if use_temp:
+                    join_conds = " AND ".join(f'main."{col}" = temp."{col}"' for col in p_unique_key_columns)
+                    select_cols = ", ".join([f'main."{col}"' for col in p_unique_key_columns] + [f'main."{pk_col}"'])
+                for i, offset in enumerate(row_iter):
+                    chunk_keys = key_df.slice(offset, chunk_size)
+                    key_rows = chunk_keys.rows()
+                    if not key_rows:
+                        continue
+                    if use_temp:
+                        # Use temp table approach
+                        temp_table_name = f"temp_keys_{i}"
+                        # Drop if exists (safety)
+                    session.execute(text(f'DROP TABLE IF EXISTS "{temp_table_name}"'))
+                    # Use UNLOGGED for faster temp table creation
+                    session.execute(text(f'CREATE TEMP TABLE "{temp_table_name}" ({col_defs_sql}) ON COMMIT DROP'))
+                    # Bulk insert all rows at once
+                    insert_stmt = f'INSERT INTO "{temp_table_name}" ({col_names_sql}) VALUES ({", ".join(param_names)})'
+                    param_dicts = [dict(zip(p_unique_key_columns, row, strict=True)) for row in key_rows]
+                    if param_dicts:
+                        session.execute(sa_text(insert_stmt), param_dicts)
+                        # Join with main table using quoted column names
+                        join_sql = f'SELECT {select_cols} FROM "{table_name}" AS main JOIN "{temp_table_name}" AS temp ON {join_conds}'
+                        results = session.execute(text(join_sql)).fetchall()
+                        # Always use "id" as the output key for the PK
+                        for row in results:
+                            row_dict = dict(zip([*p_unique_key_columns, pk_col], row, strict=True))
+                            row_dict["id"] = row_dict.pop(pk_col)
+                            ids.append(row_dict)
+                        # Drop temp table (should be auto-dropped, but explicit for safety)
+                        session.execute(text(f'DROP TABLE IF EXISTS "{temp_table_name}"'))
+                        # ...handled above for temp table join...
+                    else:
+                        # Use IN clause approach
+                        key_tuples = [tuple(row) for row in key_rows]
+                        # Get the actual PK column name and attribute
+                        pk_col = next(iter(model_class.__table__.primary_key)).name
+                        pk_attr = getattr(model_class, pk_col)
+                        query = session.query(*unique_key_attrs, pk_attr)
+                        if len(p_unique_key_columns) == 1:
+                            col_attr = unique_key_attrs[0]
+                            query = query.filter(col_attr.in_(key_tuples))
+                        else:
+                            query = query.filter(tuple_(*unique_key_attrs).in_(key_tuples))
+                        results = query.all()
+                        for row in results:
+                            row_dict = dict(zip([*p_unique_key_columns, pk_col], row, strict=True))
+                            row_dict["id"] = row_dict.pop(pk_col)
+                            ids.append(row_dict)
+                    # Manually update progress bar to show correct number of rows processed
+                    if show_progress:
+                        processed = min((i + 1) * chunk_size, total_keys)
+                        progress_bar.n = processed
+                        progress_bar.last_print_n = processed
+                        progress_bar.update(0)
+            ids_df = pl.DataFrame(ids)
+            return ids_df
+
+        if profile:
+            from line_profiler import LineProfiler
+
+            profiler = LineProfiler()
+            profiled_func = profiler(func)
+            result = profiled_func(model_class, data, chunk_size, p_unique_key_columns, table_name, use_temp_table)
+            profiler.print_stats()
+        else:
+            result = func(model_class, data, chunk_size, p_unique_key_columns, table_name, use_temp_table)
+        return result
+
     def handle_insertion_multi_line_copy(
         self,
         model_class: DeclarativeMeta,
