@@ -12,13 +12,11 @@ from foundry_sdk.db_mgmt.tables import (
     Companies,
     Datapoints,
     Dates,
-    FeatureDescriptions,
     ProductCategories,
     Products,
     Regions,
     SkuTable,
     Stores,
-    TimeSkuFeatures,
 )
 
 # from foundry_sdk.etl.constants import FlagLevels
@@ -448,7 +446,7 @@ class DataLoader:
         company_id: int,
         *,
         check_passed: bool,
-    ) -> pl.DataFrame:
+    ) -> bool:
         """
         Create and write time-sku data entries (features for each datapoint).
 
